/**
 * Netpipe - 2.0   (nocha)
 * Netpipe 2.0 is a major rewrite which features the newly designed libnpconn. more info can be
 * found in the docs/ directory. Simply put, Netpipe 2.0 is significantly more stable, much more 
 * engineered and solves many of the shortcommings that the 1.x series was susceptible to.
 *
 * Features:
 *  Uses libnpconn meaning that Netpipe supports a variable amount of protocols.
 *  Stable socket redirection which can go all the way to the packet level depending on the protocol (see the included libnpconn files)
 */

#include <npconn/conn.h>
#include <npconn/proto.h>
#include <npconn/netdef.h>

#include <stdio.h>
#include <getopt.h>

#define NAME    "Netpipe"
#define AUTHOR  "Aaron Conole"
#define VERSION "2.0"

#ifndef EXECUTE
#define EXECUTE "np"
#endif

int useage(){
  printf("%s: %s -a <target> -l <listen port> [-r <destination port>] [-s <source port>] [-i|o <protocol name>]\n", NAME, EXECUTE);
  printf("\t\t-a <target>: Required option setting the destination IP/DNS name\n"
	 "\t\t-l   <port>: Required option setting the listen port (if no -r, also defaults to destination port)\n"
	 "\t\t-r   <port>: Set the destination port to <port>\n"
	 "\t\t-s   <port>: Set the source port to <port>\n\n"
	 "\t\t-i|o <protocol>: Set the incomming or outgoing protocol to <protocol>. For a list of available protocols, see the documentation provided by the manpages or in the doc/ directory of libnpconn.\n");
  return 0;
}

int main(int argc, char *argv[]){
  int x;
  int loc, rem, src = 0;
  int verblevel = 0;
  char *host = NULL;
  char *in = NULL;
  char *out = NULL;
  char tmp[256];
  char buf[8192];
  protocol *p1, *p2;
  connection *c1, *c2, *c3;
  connection *currc;
  int isavail1, isavail2 = 0;

  netnode *n1, *n2, *n3, *n4;

  init_npconn();

  loc = rem = src;

  while ((x = getopt (argc, argv, "vha:l:s:r:o:i:")) != EOF) {
    switch(x){
    case 'h':
      return useage();
      break;
    case 'l':
      loc = atoi(optarg);
      break;
    case 's':
      src = atoi(optarg);
      break;
    case 'r':
      rem = atoi(optarg);
      break;
    case 'a':
      host = (char *)malloc(sizeof(char) * strlen(optarg));
      strcpy(host, optarg);
      break;
    case 'i':
      in = (char *)malloc(strlen(optarg));
      strcpy(in, optarg);
      break;
    case 'o':
      out = (char *)malloc(strlen(optarg));
      strcpy(out, optarg);
      break;
    case 'v':
      verblevel++;
      break;
    }
  }

  if(in == NULL){
    in = (char *)malloc(strlen("tcp-unicast") + 1);
    strcpy(in, "tcp-unicast");
  }
  p1 = get_protocol(in);

  if(out == NULL){
    out = (char *)malloc(strlen("tcp-unicast") + 1);
    strcpy(out, "tcp-unicast");
  }
  p2 = get_protocol(out);

  if(!loc){
    fprintf(stderr, "error: listening port unset or invalid\n");
    return -1;
  }

  if(host == NULL){
    fprintf(stderr, "error: host unset\n");
    return -1;
  }

  setlevel(verblevel);

  printf("Options and their values:\n"
	 "\tListen:\t%d\n"
	 "\tHost:\t%s\n", loc, host);

  sprintf(tmp, "\tDestination Port:\t%d\n"
	  "\tSource Port:\t%d\n", rem, src);
  
  print(1, tmp);

  sprintf(tmp, "\tincomming protocol:\t%s\n"
	  "\toutgoing protocol:\t%s\n", in, out);
  
  print(2, tmp);
  
  n1 = get_node(host, rem);

  if(n1 == NULL){
    print(0, "ERROR: unable to get node");
    return -1;
  }

  print(5, "DEBUG: set remote host + port\n");

  n2 = get_node(NULL, loc);

  if(n2 == NULL){
    print(0, "ERROR: unable to get node");
    return -1;
  }

  /* just to find out if there's a bug */
  n2->in.sin_port = htons(loc);

  print(5, "DEBUG: set local host + port\n");

  n3 = getlocalip(host);

  if(n3 == NULL){
    print(0, "ERROR: unable to get node\n");
    return -1;
  }

  print(5, "DEBUG: setup local out address\n");

  n3->in.sin_port = htons(src);

  print(5, "DEBUG: set local out port\n");

  n4 = get_node(NULL, 0);

  if(n4 == NULL){
    print(0, "ERROR: unable to get node\n");
    return -1;
  }

  print(5, "DEBUG: ready for connections\n");

  /* isn't this much easier ?? */
  c1 = new_connection(*n2, *n4, p1, 0);
  print(5, "DEBUG: got listen connection\n");

  if(connectin(c1, 0) < 0){
    perror("connecting");
    return -1;
  }

  print(5, "DEBUG: waiting for connection..\n");
  
  do{
    c2 = new_connection(*n3, *n1, p2, 0);
    print(5, "DEBUG: got the outgoing connection\n");
    poll_conn(c1, -1); /* just wait!! */
    print(5, "DEBUG: got a connection in\n");
    c3 = accept_new(c1);
    print(5, "Accepted connection in\n");
    currc = c3;
    bzero(buf, sizeof(buf));

    do{
      connectout(c2);
      isavail1 = poll_conn(currc, 1);
      if((isavail1 < 0)){
	disconnect(c3);
	disconnect(c2);
	continue;
      }
      if(isavail1 > 0){
	if(recvdata(currc, buf, 0, sizeof(buf)) <= 0){
	  destroy(c3);
	  destroy(c2);
	  goto out;
	}
	if(isavail2 == 0)
	  senddata(c2, buf, 0, strlen(buf));
	else
	  senddata(c3, buf, 0, strlen(buf));
      }
      if(isavail2 == 0){
	currc = c2;
	isavail2 = 1;
      }else{
	currc = c3;
	isavail2 = 0;
      }
      bzero(buf, sizeof(buf));
    }while(!isdisconnected(c3));
  out:
  }while(!isdisconnected(c1));
  return 0;
}
